classdef CanonicalHuffmanCodeBook
    
    properties (SetAccess = private)
        codeLengths;
        codeBook;
    end
    
    methods
        
        function obj = CanonicalHuffmanCodeBook(symbols, par2, val2)
            %Constructor
            %symbols,lengths antas sortert etter symbolverdi i stigende rekkeflge
            obj.codeLengths = zeros(numel(symbols), 2);
            if ( strcmpi(par2, 'lengths') )
                obj.codeLengths(:,1) = symbols;
                obj.codeLengths(:,2) = val2;
                obj.codeBook = constructCodeBook(obj);
            elseif ( strcmpi(par2, 'freqs') )
                obj.codeLengths = findLengths(obj, symbols, val2);
                obj.codeBook = constructCodeBook(obj);
            else
                error('Unknown parameter name: %s', par2);
            end
        end
                       
        function symbol = findSymbol(obj, codeword)
            %Assumes codebook sorted by codeword length (prim) and symbol (sec)
            %Uses binary search
            %Returns Inf is symbol isnt found
            
            symbol = Inf;

            if ( numel(codeword) < obj.codeBook{1}{3} ) % Check if code length is less than minimum
                return
            end
            codeLength = numel(codeword);
            %codeNumVal = bin2dec(codeword);
            codeNumVal = sum( codeword .* pow2(codeLength-1:-1:0) );
            
            minIndex = 1;
            maxIndex = numel(obj.codeBook);
            while (maxIndex >= minIndex)
                midIndex = floor( (minIndex + maxIndex) / 2 );
                if ( codeLength < obj.codeBook{midIndex}{3} )
                    maxIndex = midIndex - 1;
                elseif ( codeLength > obj.codeBook{midIndex}{3} )
                    minIndex = midIndex + 1;
                else
                    if ( codeNumVal == obj.codeBook{midIndex}{4} )
                        symbol = obj.codeBook{midIndex}{1};
                        return;
                    elseif ( codeNumVal < obj.codeBook{midIndex}{4} )
                        maxIndex = midIndex - 1;
                    else
                        minIndex = midIndex + 1;
                    end
                end
                
            end
            
        end %findSymbol
        
        function codeword = findCodeword(obj, symbol)
            for i = 1:numel(obj.codeBook)
                if ( symbol == obj.codeBook{i}{1} )
                    codeword = obj.codeBook{i}{2};
                    return;
                end
            end
        end
        
        
    end %methods
    
    methods (Access = private)        
        
        function lengths = findLengths(obj, symbols, frequencies)
            %Output:
            % - lengths: Code word lengths for all symbols (including unused symbols
            %   (length 0))
            
            huffmanCodeLengths = huffman(obj, symbols, frequencies);
            
            codeLengths = zeros( length(huffmanCodeLengths), 2 );
            
            % Find code word lengths
            for i = 1:length(huffmanCodeLengths)
                codeLengths(i,1) = huffmanCodeLengths{i}{1}; %symbol
                codeLengths(i,2) = huffmanCodeLengths{i}{2};%numel( huffmanCodeBook{i}{2} ); %length of codeword
            end
            
            % Sort by length of codeword (primary) and symbol (secondary)
            codeLengths = sortrows( codeLengths, 1); %secondary sort
            codeLengths = sortrows( codeLengths, 2); %primary sort
            
            % Construct array of (symbol, code lengths) in symbol order for all symbols
            lengths = zeros( length(symbols), 2 );
            lengths(:,1) = symbols;
            len = sortrows(codeLengths, 1);
            
            i = 1; % len index
            j = 1;  % lengths index
            while ( i <= length(len) )
                if ( lengths(j,1) == len(i,1) )
                    lengths(j,2) = len(i,2);
                    i = i + 1;
                end
                j = j + 1;
            end
            
        end %canonical_huffman
        
        function canonicalCodebook = constructCodeBook(obj)
            
            codeLengths = obj.codeLengths;
            
            % Sort by length of codeword (first) and symbol (second)
            codeLengths = sortrows( codeLengths, 2);
            
            %remove symbols with zero freq
            firstNonzeroIndex = length(codeLengths) - nnz(codeLengths(:,2)) + 1;
            l = zeros( nnz(codeLengths(:,2)), 2 );
            l(:,1) = codeLengths( firstNonzeroIndex: length(codeLengths), 1);
            l(:,2) = codeLengths( firstNonzeroIndex: length(codeLengths), 2);

            canonicalCodebook = cell(length(l), 1);
            
            % Replace code words
            canonicalCodebook{1}{1} = l(1,1); %symbol
            canonicalCodebook{1}{2} = zeros(1, l(1,2)); %codeword
            canonicalCodebook{1}{3} = l(1,2); %codelength
            canonicalCodebook{1}{4} = 0; %decimal value of codeword
            for i = 2:length(canonicalCodebook)
                canonicalCodebook{i}{1} = l(i,1); %symbol

                tempVal = canonicalCodebook{i-1}{4} + 1;
                [f, e] = log2(tempVal);
                tempLength = max( e, canonicalCodebook{i-1}{3} );
                
                newWord = zeros(1, l(i,2));
                newWord(1:tempLength) = bitget( tempVal, tempLength:-1:1 );

                canonicalCodebook{i}{2} = newWord; %codeword
                canonicalCodebook{i}{3} = l(i,2); %codelength
                canonicalCodebook{i}{4} = sum( newWord .* pow2(l(i,2)-1:-1:0) ); %decimal value of codeword
            end

        end
                        
        
        
        function codeLengths = huffman(obj, symbols, frequencies)
            % Finds length of Huffman code words. Does NOT return any code
            % words, just what their lengths would be.
            %
            % Disclaimer: Based on code from "Digital Image Processing Using Matlab" (adapted)
            
            % Remove symbols with zero freq
            [frequencies ind] = sort(frequencies);
            symbols = symbols(ind);
            
            firstNonzeroIndex = length(frequencies) - nnz(frequencies) + 1;
            frequencies = frequencies( firstNonzeroIndex:length(frequencies) );
            symbols = symbols( firstNonzeroIndex:length(symbols) );
            
            % Return if only one symbol is used
            if (length(frequencies) == 1)
                codeLengths = { symbols, '1' };
                return
            end
            
            % Build Huffman tree
            tree = cell( 1, length(symbols) );
            for i = 1:length(symbols)
                tree{i} = i;   % i is used instead of symbol value in case of negative symbols. Mapped back later.
            end
            
            while (numel(tree) > 2)
                [frequencies, ind] = sort(frequencies);
                tree = tree(ind);
                frequencies(2) = frequencies(1) + frequencies(2);
                frequencies(1) = [];
                tree{2} = { tree{1}, tree{2} };
                tree(1) = [];
            end
            
            % Create codebook (get lengths from traverseHuffmanTree and
            % map back symbols
            code = traverseHuffmanTree(obj, tree, {}, 0);
            codeLengths = cell( numel(code), 1 );
            for i = 1: numel(code)
                codeLengths{i} = { symbols(i), code{i} };
            end
            
        end %huffman
        
        function code = traverseHuffmanTree(obj, huffmanTree, subCode, level)
            % Disclaimer: Based on code from "Digital Image Processing Using Matlab" (adapted)
            code = subCode;
            if ( isa(huffmanTree, 'cell') ) % huffmanTree is a subtree
                code = traverseHuffmanTree(obj, huffmanTree{1}, code, level + 1);
                code = traverseHuffmanTree(obj, huffmanTree{2}, code, level + 1);
            else    % huffmanTree is a leaf node
                code{huffmanTree} = level;
            end
            
        end
        
    end %methods
    
end