classdef Decoder < handle
    
    properties (SetAccess = private) % I/O properties
        inputFolder;
        outputFolder;
        outputFileName;
        originalFileName;
        yuvWriter;
        logFile;
    end
    
    properties (SetAccess = private) % Props read from stream header
        numOfTemporalLevels;
        filterType;
        lumaBlockSize;
        frameWidth;
        frameHeight;
        frameRate;
        numOfFrames;
        mvEntropyCoder;
        numOfHuffmanSymbols;
    end
    
    properties  (SetAccess = private) % Derived props
        chromaBlockSize;
    end
    
    methods % Public methods
        
        function obj = Decoder(inputFolder, outputFolder, outputFileName, logFile, varargin)
            % Constructor
            obj.inputFolder = inputFolder;
            obj.outputFolder = outputFolder;
            obj.outputFileName = outputFileName;
            obj.logFile = logFile;
            
            if (nargin == 5)
                obj.originalFileName = varargin{1};
            end            
            
            if ( ~exist(obj.outputFolder, 'file') )
               mkdir(outputFolder);
           end
            
            readHeader(obj);
            
            obj.chromaBlockSize = obj.lumaBlockSize / 2;
            
            obj.yuvWriter = YUVWriter([obj.outputFolder '/' obj.outputFileName], obj.frameWidth, obj.frameHeight);
        end
        
        function decode(obj)
            startTime = tic;
            if ( strcmp(obj.filterType, 'haar') )
                haarDecoding(obj);
            elseif (strcmp(obj.filterType, '1/3') )
                oneThreeDecoding(obj);
            end
            log(obj, startTime)
        end
        
        function delete(obj)
            % Destructor
            delete(obj.yuvWriter);
        end
        
    end
    
    methods (Access = private)
        
        function readHeader(obj)
            fileID = fopen([obj.inputFolder '/header.bin']);
            
            % Read number of temporal levels
            obj.numOfTemporalLevels = fread( fileID, 1, ['ubit' num2str(ConstantProperties.MAX_NO_TEMPORAL_LEVELS_LOG2)] );
            
            % Read filter type
            filter = fread(fileID, 1, 'ubit1');
            obj.filterType = '';
            if ( filter == 0 )
                obj.filterType = 'haar';
            else
                obj.filterType = '1/3';
            end            
            
            % Read luma blocksize
            obj.lumaBlockSize = fread( fileID, 1, ['ubit' num2str(ConstantProperties.MAX_LUMA_BLOCKSIZE_LOG2)] );
            
            % Read frame width
            obj.frameWidth = fread( fileID, 1, ['ubit' num2str(ConstantProperties.MAX_FRAME_WIDTH_LOG2)] );
            
            % Read frame height
            obj.frameHeight = fread( fileID, 1, ['ubit' num2str(ConstantProperties.MAX_FRAME_HEIGHT_LOG2)] );
            
            % Read frame rate
            obj.frameRate = fread( fileID, 1, ['ubit' num2str(ConstantProperties.MAX_FRAME_RATE_LOG2)] );
            
            % Read number of encoded frames
            obj.numOfFrames = fread(fileID, 1, 'ubit20');
            
            % Read motion vector entropy coder type
            obj.mvEntropyCoder = fread(fileID, 1, 'ubit1');
            
            % Read number of Huffman symbols in MV codebooks
            obj.numOfHuffmanSymbols = fread( fileID, 1, ['ubit' num2str(ConstantProperties.MAX_NO_HUFFMAN_SYMBOLS_LOG2)] );
            
            % Close file
            fclose(fileID);
        end
        
        function haarDecoding(obj)
            %GOP-wise decoding
            remainingFrames = obj.numOfFrames;
            currentGOP = 1;
            tLevels = obj.numOfTemporalLevels;
            while (remainingFrames > 1)
                while(tLevels > 0)
                    if (2^tLevels > remainingFrames)
                        tLevels = tLevels - 1;
                    else
                        break;
                    end
                end
                currentGOPsize = 2^tLevels;
                % Read /decode jp2k GOP (returns decoded mctf subbands)
                [luma_sub_rec, Cb_sub_rec, Cr_sub_rec] = jp2k_gop_decoder(currentGOP, currentGOPsize, obj.inputFolder, obj.outputFolder, obj.frameWidth, obj.frameHeight);
                remainingFrames = remainingFrames - currentGOPsize;
                
                % Read / decode vectors
                [luma_mvx_rec, luma_mvy_rec, Cb_mvx_rec, Cb_mvy_rec, Cr_mvx_rec, Cr_mvy_rec] = motion_vector_gop_decoder(currentGOP, currentGOPsize-1, obj.frameWidth, obj.frameHeight, obj.lumaBlockSize, obj.inputFolder, obj.mvEntropyCoder, obj.numOfHuffmanSymbols);
                
                % MCTF synthesis
                luma_rec = mctf_gop_decoder(luma_sub_rec, luma_mvx_rec, luma_mvy_rec, tLevels, obj.filterType, obj.lumaBlockSize);
                Cb_rec = mctf_gop_decoder(Cb_sub_rec, Cb_mvx_rec, Cb_mvy_rec, tLevels, obj.filterType, obj.chromaBlockSize);
                Cr_rec = mctf_gop_decoder(Cr_sub_rec, Cr_mvx_rec, Cr_mvy_rec, tLevels, obj.filterType, obj.chromaBlockSize);
                
                % Write reconstructed frames to yuv file
                writeNextFrames(obj.yuvWriter, luma_rec, Cb_rec, Cr_rec);
                
                currentGOP = currentGOP + 1;
            end
            % If there is one frame left, read as IDR
            if (remainingFrames == 1)
                [luma_lastFrame, Cb_lastFrame, Cr_lastFrame] = jp2k_gop_decoder(currentGOP, 1, obj.inputFolder, obj.outputFolder, obj.frameWidth, obj.frameHeight);
                writeNextFrame(obj.yuvWriter, uint8(luma_lastFrame), uint8(Cb_lastFrame), uint8(Cr_lastFrame));
            end
        end
        
        function oneThreeDecoding(obj)
            %GOP-wise decoding
            remainingFrames = obj.numOfFrames;
            
            % 1st frame (GOP) is IDR
            [luma_idr, Cb_idr, Cr_idr] = jp2k_gop_decoder(1, 1, obj.inputFolder, obj.outputFolder, obj.frameWidth, obj.frameHeight);
            writeNextFrame(obj.yuvWriter, uint8(luma_idr), uint8(Cb_idr), uint8(Cr_idr));
            remainingFrames = remainingFrames - 1;
            
            currentGOP = 2;
            tLevels = obj.numOfTemporalLevels;
            while (remainingFrames > 1)
                while(tLevels > 0)
                    if (2^tLevels > remainingFrames)
                        tLevels = tLevels - 1;
                    else
                        break;
                    end
                end
                currentGOPsize = 2^tLevels;
                % Read /decode jp2k GOP (returns decoded mctf subbands)
                [luma_sub_rec, Cb_sub_rec, Cr_sub_rec] = jp2k_gop_decoder(currentGOP, currentGOPsize, obj.inputFolder, obj.outputFolder, obj.frameWidth, obj.frameHeight);
                remainingFrames = remainingFrames - currentGOPsize;
                
                % Read / decode vectors
                [luma_mvx_rec, luma_mvy_rec, Cb_mvx_rec, Cb_mvy_rec, Cr_mvx_rec, Cr_mvy_rec] = motion_vector_gop_decoder(currentGOP, 2^(tLevels+1) - 2, obj.frameWidth, obj.frameHeight, obj.lumaBlockSize, obj.inputFolder, obj.mvEntropyCoder, obj.numOfHuffmanSymbols);
                
                % MCTF synthesis
                luma_rec = mctf_gop_decoder(luma_sub_rec, luma_mvx_rec, luma_mvy_rec, tLevels, obj.filterType, obj.lumaBlockSize, luma_idr);
                Cb_rec = mctf_gop_decoder(Cb_sub_rec, Cb_mvx_rec, Cb_mvy_rec, tLevels, obj.filterType, obj.chromaBlockSize, Cb_idr, 0);
                Cr_rec = mctf_gop_decoder(Cr_sub_rec, Cr_mvx_rec, Cr_mvy_rec, tLevels, obj.filterType, obj.chromaBlockSize, Cr_idr, 0);
                
                % Write reconstructed frames to yuv file
                writeNextFrames(obj.yuvWriter, luma_rec, Cb_rec, Cr_rec);
                
                % Lagre siste frame i GOP som IDR
                luma_idr = luma_rec(:,:,currentGOPsize);
                Cb_idr = Cb_rec(:,:,currentGOPsize);
                Cr_idr = Cr_rec(:,:,currentGOPsize);
                
                currentGOP = currentGOP + 1;
            end
            % If there is one frame left, read as IDR
            if (remainingFrames == 1)
                [luma_lastFrame, Cb_lastFrame, Cr_lastFrame] = jp2k_gop_decoder(currentGOP, 1, obj.inputFolder, obj.outputFolder, obj.frameWidth, obj.frameHeight);
                writeNextFrame(obj.yuvWriter, uint8(luma_lastFrame), uint8(Cb_lastFrame), uint8(Cr_lastFrame));
            end
        end
        
        function log(obj, startTime)
            fileID = 1;
            if (~isempty(obj.logFile))
                fileID = fopen(obj.logFile, 'at');
            end
            fprintf(fileID, '\nProposed Wavelet Codec for master thesis by F.T. Erlid, IET, NTNU, 2012\n');
            fprintf(fileID, '\nDecoder Log\n');
            
            fprintf(fileID, '\nI/O \n');
            fprintf(fileID, 'Input stream folder       : %s \n', obj.inputFolder);
            fprintf(fileID, 'Decoded sequence          : %s \n', obj.outputFileName);
            fprintf(fileID, 'Sequence length           : %d \n', obj.numOfFrames);
            fprintf(fileID, 'Input Video format        : %dx%d @ %d fps \n', obj.frameWidth, obj.frameHeight, obj.frameRate);
            %fprintf('Output Video format       : %dx%d @ x fps \n', obj.frameWidth, obj.frameHeight);
            
            fprintf(fileID, '\nMotion Estimation (Full-pixel accuracy) \n');
            fprintf(fileID, 'Luma Block size           : %d \n', obj.lumaBlockSize);
            fprintf(fileID, 'Chroma Block size         : %d \n', obj.chromaBlockSize);
            
            fprintf(fileID, '\nMCTF \n');
            fprintf(fileID, 'MCTF Filter type          : %s \n', obj.filterType);
            fprintf(fileID, 'Number of temporal levels : %d \n', obj.numOfTemporalLevels);
            fprintf(fileID, 'GOP size                  : %d \n', 2^obj.numOfTemporalLevels);
                        
            
            % Metrics calculations
            if ( ~isempty(obj.originalFileName) )
                origReader = YUVReader(obj.originalFileName, obj.frameWidth, obj.frameHeight, 0);
                recReader = YUVReader([obj.outputFolder '/' obj.outputFileName], obj.frameWidth, obj.frameHeight, 0);                
                
                fprintf(fileID, '\n      |                PSNR                  |  ');
                fprintf(fileID, '              SSIM              \n');
                fprintf(fileID, 'Frame |  Y-PSNR       U-PSNR       V-PSNR    |  ');
                fprintf(fileID, 'Y-SSIM       U-SSIM       V-SSIM\n');
                fprintf(fileID, '--------------------------------------------------');
                fprintf(fileID, '--------------------------------\n');
                
                avgPSNR = [0, 0, 0];
                avgSSIM = [0, 0, 0];
                
                frame = 1;
                while ( hasNextFrame(recReader) )
                    %Read frames
                    [origFrame_luma, origFrame_Cb, origFrame_Cr] = readNextFrame(origReader);
                    [recFrame_luma, recFrame_Cb, recFrame_Cr] = readNextFrame(recReader);                    
                    
                    if (numel(recFrame_luma) == 0)
                        continue;
                    end
                    
                    %Preprocessing for speedup purposes
                    origFrame_luma = preprocess_metrix_mux(origFrame_luma);
                    origFrame_Cb = preprocess_metrix_mux(origFrame_Cb);
                    origFrame_Cr = preprocess_metrix_mux(origFrame_Cr);
                    recFrame_luma = preprocess_metrix_mux(recFrame_luma);
                    recFrame_Cb = preprocess_metrix_mux(recFrame_Cb);
                    recFrame_Cr = preprocess_metrix_mux(recFrame_Cr);
                    
                    % PSNR calculations
                    yPSNR = metrix_mux(origFrame_luma, recFrame_luma, 'psnr');
                    uPSNR = metrix_mux(origFrame_Cb, recFrame_Cb, 'psnr');
                    vPSNR = metrix_mux(origFrame_Cr, recFrame_Cr, 'psnr');
                    
                    avgPSNR = avgPSNR + [yPSNR, uPSNR, vPSNR];
                    
                    % SSIM calculations
                    ySSIM = metrix_mux(origFrame_luma, recFrame_luma, 'ssim');
                    uSSIM = metrix_mux(origFrame_Cb, recFrame_Cb, 'ssim');
                    vSSIM = metrix_mux(origFrame_Cr, recFrame_Cr, 'ssim');
                    
                    avgSSIM = avgSSIM + [ySSIM, uSSIM, vSSIM];
                    
                    fprintf(fileID, '%-4d  |  %-8.4f     %-8.4f     %-8.4f  |  %-8.4f     %-8.4f     %-8.4f \n', frame, yPSNR, uPSNR, vPSNR, ySSIM, uSSIM, vSSIM);
                    frame = frame + 1;
                end     
                avgPSNR = avgPSNR / obj.numOfFrames;
                avgSSIM = avgSSIM / obj.numOfFrames;
                fprintf(fileID, '--------------------------------------------------');
                fprintf(fileID, '--------------------------------\n');
                fprintf(fileID, 'Avg.  |  %-8.4f     %-8.4f     %-8.4f  |  %-8.4f     %-8.4f     %-8.4f \n', avgPSNR(1), avgPSNR(2), avgPSNR(3), avgSSIM(1), avgSSIM(2), avgSSIM(3));
            end
            
            fprintf(fileID, '\nTotal decoding time: %.3f sec. \n\n', toc(startTime));
            
            fclose(fileID);
        end
        
    end
    
end