classdef Encoder
    % *********** Info ***********
    %
    % - T+2D wavelet encoder
    % - Method of Motion Estimation: Block Matching (full-pixel accuracy only).
    %   Cost function: MSE (Mean Squared Error)
    % - Motion Estimation search type: Diamond Search or Full Search.
    %   Search range is doubled for every lower temporal level.
    %   lumaSearchRange parameter gives maximal search range (for luma).
    % - Method of Motion Vector encoding: Canonical Huffman encoding or
    %   Arithmetic coding
    % - MCTF filter type: Haar or 1/3 filter
    % - 2D coder: JPEG 2000 - JasPer 1.900.1.
    %
    % Author: F. T. Erlid
    
   properties (SetAccess = private) % Input properties
       fileName;            % Name of YUV4:2:0 file to encode
       firstFrame;          % First frame to encode (0-indexed)
       lastFrame;           % Last frame to encode (0-indexed)
       frameWidth;          % Frame width
       frameHeight;         % Frame height
       frameRate;           % Frame rate
       lumaBlockSize;       % Block size for Motion Estimation (Luma, halved for chroma components)
       lumaSearchRange;     % Max Search range for Motion Estimation (Luma, halved for chroma components)
       searchType;          % Motion Estimation search type
       filterType;          % Filter type for MCTF (Haar or 1/3 filter)
       numOfTemporalLevels; % Number of temporal levels for MCTF decomposition
       targetRate;          % Target output bitrate [kbps] (0 = lossless JP2k coding)
       mvEntropyCoder;      % 0: Canonical Huffman 1: Arithmetic coding
       outputFolder;        % Name of output folder
       logFile;             % Name of encoder log file (optional).
       lumaBudgetPercent;
       CbBudgetPercent;
       lpPercent;
   end
   
   properties (SetAccess = private) % Derived parameters
       numOfFrames;         % Number of frames to encode
       chromaBlockSize;     % Blocksize for chroma ME
       chromaSearchRange;   % SearchRange for chroma ME
       yuvReader;           % Instance of YUVReader
       totalByteBudget;     % Total byte budget for rate control
   end
   
   methods %Public methods
       
       function obj = Encoder(fileName, firstFrame, lastFrame, frameWidth, frameHeight,...
               frameRate, lumaBlockSize, lumaSearchRange, searchType, filterType,...
               numOfTemporalLevels, targetRate, mvEntropyCoder, outputFolder, lumaBudgetPercent, CbBudgetPercent, lpPercent, varargin)
           % Constructor
           
           obj.fileName = fileName;
           obj.firstFrame = firstFrame;
           obj.lastFrame = lastFrame;
           obj.frameWidth = frameWidth;
           obj.frameHeight = frameHeight;
           obj.frameRate = frameRate;
           obj.lumaBlockSize = lumaBlockSize;
           obj.lumaSearchRange = lumaSearchRange;
           obj.searchType = searchType;
           obj.filterType = filterType;
           obj.numOfTemporalLevels = numOfTemporalLevels;
           obj.targetRate = targetRate;
           obj.mvEntropyCoder = mvEntropyCoder;
           obj.outputFolder = outputFolder;
           
           obj.lumaBudgetPercent = lumaBudgetPercent;
           obj.CbBudgetPercent = CbBudgetPercent;
           obj.lpPercent = lpPercent;
           
           if (nargin == 18)
               obj.logFile = varargin{1};
           end
           
           obj.numOfFrames = obj.lastFrame - obj.firstFrame + 1;
           obj.chromaBlockSize = obj.lumaBlockSize / 2;
           obj.chromaSearchRange = obj.lumaSearchRange / 2;
           
           obj.totalByteBudget = ((obj.targetRate*1000/8) / obj.frameRate) * obj.numOfFrames; % bytes/frame * number of frames
             
           validateProperties(obj);
           
           if ( ~exist(obj.outputFolder, 'file') )
               mkdir(outputFolder);
           end
           
           obj.yuvReader = YUVReader(obj.fileName, obj.frameWidth, obj.frameHeight, obj.firstFrame);
           
       end
       
       function encode(obj)
           startTime = tic;
           
           delete( [obj.outputFolder '/*'] ); % Remove old files from stream folder           
           obj.totalByteBudget = obj.totalByteBudget - writeHeader(obj);

           if ( strcmp(obj.filterType, 'haar') )
               haarEncoding(obj);
           elseif ( strcmp(obj.filterType, '1/3') )
               oneThreeEncoding(obj);
           end
           log(obj, startTime);
       end
      
       function delete(obj)
           % Destructor
           delete(obj.yuvReader);
       end
       
   end
   
   methods (Access = private)
       
       function totalBytesWritten = writeHeader(obj)         
           if ( strcmpi(obj.filterType, 'haar') )
               filter = 0;
           elseif ( strcmpi(obj.filterType, '1/3') )
               filter = 1;
           else
               error('Unknown filter type: %s', obj.filterType);
           end
           
           totalBitsWritten = 0;
           
           fileID = fopen([obj.outputFolder '/header.bin'], 'w');
           
           % Write num. of temporal levels with [ConstantProperties.MAX_NO_TEMPORAL_LEVELS_LOG2] bits
           totalBitsWritten = totalBitsWritten + ConstantProperties.MAX_NO_TEMPORAL_LEVELS_LOG2 * fwrite( fileID, obj.numOfTemporalLevels, ['ubit' num2str(ConstantProperties.MAX_NO_TEMPORAL_LEVELS_LOG2)] );
           
           % Write filter type with 1 bit (0 for haar filter, 1 for 1/3 filter)
           totalBitsWritten = totalBitsWritten + fwrite(fileID, filter, 'ubit1');           
           
           % Write block matching ME block size for luma with [ConstantProperties.MAX_LUMA_BLOCKSIZE_LOG2] bits.
           totalBitsWritten = totalBitsWritten + ConstantProperties.MAX_LUMA_BLOCKSIZE_LOG2 * fwrite( fileID, obj.lumaBlockSize, ['ubit' num2str(ConstantProperties.MAX_LUMA_BLOCKSIZE_LOG2)] );
           
           % Write frame width with [ConstantProperties.MAX_FRAME_WIDTH_LOG2] bits.
           totalBitsWritten = totalBitsWritten + ConstantProperties.MAX_FRAME_WIDTH_LOG2 * fwrite( fileID, obj.frameWidth, ['ubit' num2str(ConstantProperties.MAX_FRAME_WIDTH_LOG2)] );
           
           % Write frame height with [MAX_FRAME_HEIGHT_LOG2] bits.
           totalBitsWritten = totalBitsWritten + ConstantProperties.MAX_FRAME_HEIGHT_LOG2 * fwrite( fileID, obj.frameHeight, ['ubit' num2str(ConstantProperties.MAX_FRAME_HEIGHT_LOG2)] );
           
           % Write frame rate with [MAX_FRAME_RATE_LOG2] bits
           totalBitsWritten = totalBitsWritten + ConstantProperties.MAX_FRAME_RATE_LOG2 * fwrite( fileID, obj.frameRate, ['ubit' num2str(ConstantProperties.MAX_FRAME_RATE_LOG2)] );
           
           % Write number of encoded frames with 20 bits.
           totalBitsWritten = totalBitsWritten + 20 * fwrite(fileID, obj.numOfFrames, 'ubit20');
           
           % Write type of motion vector entropy coder
           totalBitsWritten = totalBitsWritten + fwrite(fileID, obj.mvEntropyCoder, 'ubit1');
           
           % Write number of Huffman symbols for encoded MV data      
           noHuffmanSymbols = (2 * obj.lumaSearchRange) - 1;
           totalBitsWritten = totalBitsWritten + ConstantProperties.MAX_NO_HUFFMAN_SYMBOLS_LOG2 * fwrite( fileID, noHuffmanSymbols, ['ubit' num2str(ConstantProperties.MAX_NO_HUFFMAN_SYMBOLS_LOG2)] );
           
           totalBytesWritten = ceil( totalBitsWritten / 8 );
           fclose(fileID);
       end
       
       function haarEncoding(obj)
           %Main loop (GOP-wise encoding)
           remainingFrames = obj.numOfFrames;
           currentGOP = 1;
           tLevels = obj.numOfTemporalLevels;
           while(remainingFrames > 1)
               while(tLevels > 0)
                   if (2^tLevels > remainingFrames)
                       tLevels = tLevels - 1;
                   else
                       break;
                   end
               end
               currentGOPsize = 2^tLevels;
               
               % Read GOP
               [luma_gop, Cb_gop, Cr_gop] = readNextFrames(obj.yuvReader, currentGOPsize);               
               
               % MCTF
               [luma_subbands, luma_mvx, luma_mvy] = mctf_gop_encoder(luma_gop, tLevels, obj.filterType, obj.lumaBlockSize, obj.lumaSearchRange, obj.searchType);
               [Cb_subbands, Cb_mvx, Cb_mvy] = mctf_gop_encoder(Cb_gop, tLevels, obj.filterType, obj.chromaBlockSize, obj.chromaSearchRange, obj.searchType);
               [Cr_subbands, Cr_mvx, Cr_mvy] = mctf_gop_encoder(Cr_gop, tLevels, obj.filterType, obj.chromaBlockSize, obj.chromaSearchRange, obj.searchType);
               
               gopByteBudget = (obj.totalByteBudget / remainingFrames ) * currentGOPsize;
               
               % Komprimer vektorer (lossless)
               mvBytes = motion_vector_gop_encoder(luma_mvx, luma_mvy, Cb_mvx, Cb_mvy, Cr_mvx, Cr_mvy, obj.lumaSearchRange, currentGOP, obj.outputFolder, obj.mvEntropyCoder);             
               obj.totalByteBudget = obj.totalByteBudget - mvBytes;
               gopByteBudget = gopByteBudget - mvBytes;
               
               % Lagre subband frames som jp2k
               jp2kBytes = jp2k_gop_encoder(luma_subbands, Cb_subbands, Cr_subbands, currentGOP, gopByteBudget, obj.outputFolder);
               obj.totalByteBudget = obj.totalByteBudget - jp2kBytes;
               
               currentGOP = currentGOP + 1;
               remainingFrames = remainingFrames - currentGOPsize;
           end
           % If there is one frame left, store as IDR
           if (remainingFrames == 1)
               [luma_idr, Cb_idr, Cr_idr] = readNextFrame(obj.yuvReader);
               jp2k_gop_encoder(luma_idr, Cb_idr, Cr_idr, currentGOP, obj.targetRate, obj.outputFolder);
           end
       end
       
       function oneThreeEncoding(obj)           
           %Main loop (GOP-wise encoding)
           remainingFrames = obj.numOfFrames;
           
           % 1st frame (GOP) is IDR
           frameByteBudget = (obj.totalByteBudget / remainingFrames )*1.5;
           [luma_idr, Cb_idr, Cr_idr] = readNextFrame(obj.yuvReader);
           jp2kBytes = jp2k_gop_encoder(luma_idr, Cb_idr, Cr_idr, 1, frameByteBudget, obj.outputFolder, obj.lumaBudgetPercent, obj.CbBudgetPercent, obj.lpPercent);
           obj.totalByteBudget = obj.totalByteBudget - jp2kBytes;
           remainingFrames = remainingFrames - 1;
           
           currentGOP = 2;
           tLevels = obj.numOfTemporalLevels;
           while(remainingFrames > 1)
               while(tLevels > 0)
                   if (2^tLevels > remainingFrames)
                       tLevels = tLevels - 1;
                   else
                       break;
                   end
               end
               currentGOPsize = 2^tLevels;
               
               % Read GOP
               [luma_gop, Cb_gop, Cr_gop] = readNextFrames(obj.yuvReader, currentGOPsize);
               
               % MCTF
               [luma_subbands, luma_mvx, luma_mvy] = mctf_gop_encoder(luma_gop, tLevels, obj.filterType, obj.lumaBlockSize, obj.lumaSearchRange, obj.searchType, luma_idr);
               [Cb_subbands, Cb_mvx, Cb_mvy] = mctf_gop_encoder(Cb_gop, tLevels, obj.filterType, obj.chromaBlockSize, obj.chromaSearchRange, obj.searchType, Cb_idr, 0);
               [Cr_subbands, Cr_mvx, Cr_mvy] = mctf_gop_encoder(Cr_gop, tLevels, obj.filterType, obj.chromaBlockSize, obj.chromaSearchRange, obj.searchType, Cr_idr, 0);
               
               gopByteBudget = (obj.totalByteBudget / remainingFrames ) * currentGOPsize;
               
               % Komprimer vektorer (lossless)
               mvBytes = motion_vector_gop_encoder(luma_mvx, luma_mvy, Cb_mvx, Cb_mvy, Cr_mvx, Cr_mvy, obj.lumaSearchRange, currentGOP, obj.outputFolder, obj.mvEntropyCoder);
               obj.totalByteBudget = obj.totalByteBudget - mvBytes;
               gopByteBudget = gopByteBudget - mvBytes;
               
               % Lagre subband frames som jp2k
               jp2kBytes = jp2k_gop_encoder(luma_subbands, Cb_subbands, Cr_subbands, currentGOP, gopByteBudget, obj.outputFolder, obj.lumaBudgetPercent, obj.CbBudgetPercent, obj.lpPercent);
               obj.totalByteBudget = obj.totalByteBudget - jp2kBytes;
               
               currentGOP = currentGOP + 1;
               
               %Lagre siste frame i GOP som IDR
               luma_idr = luma_gop(:,:,currentGOPsize);
               Cb_idr = Cb_gop(:,:,currentGOPsize);
               Cr_idr = Cr_gop(:,:,currentGOPsize);
               
               remainingFrames = remainingFrames - currentGOPsize;
           end
           % If there is one frame left, store as IDR
           if (remainingFrames == 1)
               [luma_idr, Cb_idr, Cr_idr] = readNextFrame(obj.yuvReader);
               frameByteBudget = (obj.totalByteBudget / remainingFrames );
               jp2k_gop_encoder(luma_idr, Cb_idr, Cr_idr, currentGOP, frameByteBudget, obj.outputFolder, obj.lumaBudgetPercent, obj.CbBudgetPercent, obj.lpPercent);
           end
       end
       
       function validateProperties(obj)
           %Filename is checked in YUVReader
           % Checks first frame
           if (obj.firstFrame < 0 || rem(obj.firstFrame, 1) ~= 0 )
               error('firstFrame parameter must be integer larger than 0');
           end
           % Checks last frame
           if (obj.lastFrame < obj.firstFrame ||rem(obj.lastFrame, 1) ~= 0 )
               error('lastFrame parameter must be integer not less than firstFrame parameter');
           end
           % Checks numOfFrames
           if ( obj.numOfFrames < 1 )
               error('No frames to encode');
           end
           % Checks frameWidth
           if ( obj.frameWidth < ConstantProperties.MIN_FRAME_WIDH || obj.frameWidth > ConstantProperties.MAX_FRAME_WIDTH || rem(obj.frameWidth, 1) ~= 0 )
               error( 'frameWidth parameter must be an integer in range [%d, %d]', ConstantProperties.MIN_FRAME_WIDH, ConstantProperties.MAX_FRAME_WIDH );
           end
           % Checks frameHeight
           if ( obj.frameHeight < ConstantProperties.MIN_FRAME_HEIGHT || obj.frameHeight > ConstantProperties.MAX_FRAME_HEIGHT || rem(obj.frameHeight, 1) ~= 0 )
               error( 'frameHeight parameter must be an integer in range [%d, %d]', ConstantProperties.MIN_FRAME_HEIGHT, ConstantProperties.MAX_FRAME_HEIGHT );
           end
           % Checks frame rate
           if ( obj.frameRate < ConstantProperties.MIN_FRAME_RATE || obj.frameRate > ConstantProperties.MAX_FRAME_RATE || rem(obj.frameRate, 1) ~= 0 )
               error('frameRate parameter must be an integer in range [%d, %d]', ConstantProperties.MIN_FRAME_RATE, ConstantProperties.MAX_FRAME_RATE);
           end           
           % Checks lumaBlockSize
           if ( obj.lumaBlockSize < ConstantProperties.MIN_LUMA_BLOCKSIZE || obj.lumaBlockSize > ConstantProperties.MAX_LUMA_BLOCKSIZE || rem(obj.lumaBlockSize, 1) ~= 0 )
               error( 'lumaBlockSize parameter must be an integer in range [%d, %d]', ConstantProperties.MIN_LUMA_BLOCKSIZE, ConstantProperties.MAX_LUMA_BLOCKSIZE );
           elseif ( mod(obj.lumaBlockSize, 2) )
               error('Block size must be even');
           end
           % Checks searchRange
           if ( obj.lumaSearchRange < ConstantProperties.MIN_LUMA_SEARCHRANGE || obj.lumaSearchRange > ConstantProperties.MAX_LUMA_SEARCHRANGE || rem(obj.lumaSearchRange, 1) ~= 0 )
               error('lumaSearchRange parameter must be an integer in range [%d, %d]', ConstantProperties.MIN_LUMA_SEARCHRANGE, ConstantProperties.MAX_LUMA_SEARCHRANGE);
           elseif ( mod(obj.lumaSearchRange, 2) )
               error('Search range must be even');
           end           
           % Checks searthType
           if ( ~sum( strcmpi(obj.searchType, ConstantProperties.SUPPORTED_SEARCH_TYPES) ) )
               error('Unknown search type: %s', obj.searchType);
           end
           % Checks filterType
           if ( ~sum( strcmpi(obj.filterType, ConstantProperties.SUPPORTED_FILTER_TYPES) ) )
               error('Unknown filter type: %s', obj.filterType);
           end
           % Checks numOfTemporalLevels
           if ( obj.numOfTemporalLevels < ConstantProperties.MIN_NO_TEMPORAL_LEVELS || obj.numOfTemporalLevels > ConstantProperties.MAX_NO_TEMPORAL_LEVELS || rem(obj.numOfTemporalLevels, 1) ~= 0 )
               error( 'numOfTemporalLevels parameter must be an integer in range [%d, %d]', ConstantProperties.MIN_NO_TEMPORAL_LEVELS, ConstantProperties.MAX_NO_TEMPORAL_LEVELS );
           end           
           %Checks mvEntropyCoder
           if (obj.mvEntropyCoder ~= 0 && obj.mvEntropyCoder ~= 1)
               error('Parameter mvEntropyCoder must be either 0 or 1');
           end
           
       end
       
       function log(obj, startTime)
           fileID = 1;
           if (~isempty(obj.logFile))
               fileID = fopen(obj.logFile, 'wt');
           end
           fprintf(fileID, '\nProposed Wavelet Codec for master thesis by F.T. Erlid, IET, NTNU, 2012\n');
           fprintf(fileID, '\nEncoder Log\n');
                                 
           fprintf(fileID, '\nI/O \n');
           fprintf(fileID, 'Input file                : %s \n', obj.fileName);
           fprintf(fileID, 'Output stream folder      : %s \n', obj.outputFolder);
           fprintf(fileID, 'Video format              : %dx%d @ %d fps \n', obj.frameWidth, obj.frameHeight, obj.frameRate);
           fprintf(fileID, 'Encoded frames            : %d - %d (%d frames) \n', obj.firstFrame, obj.lastFrame, obj.numOfFrames);
           fprintf(fileID, 'Target rate               : %d kbps\n', obj.targetRate);
           
           fprintf(fileID, '\nMotion Estimation (Full-pixel accuracy) \n');
           fprintf(fileID, 'Luma Block size           : %d \n', obj.lumaBlockSize);
           fprintf(fileID, 'Chroma Block size         : %d \n', obj.chromaBlockSize);
           fprintf(fileID, 'Luma Search range         : %d \n', obj.lumaSearchRange);
           fprintf(fileID, 'Chroma Search range       : %d \n', obj.chromaSearchRange);
           fprintf(fileID, 'Search type               : %s \n', obj.searchType);
           
           fprintf(fileID, '\nMCTF \n');
           fprintf(fileID, 'Filter type               : %s \n', obj.filterType);           
           fprintf(fileID, 'Number of temporal levels : %d \n', obj.numOfTemporalLevels);
           fprintf(fileID, 'GOP size                  : %d \n', 2^obj.numOfTemporalLevels);
           fprintf(fileID, '\n\n');
           
           fprintf(fileID, 'Luma byte budget percent  : %.3f \n', obj.lumaBudgetPercent);
           fprintf(fileID, 'Cb byte budget percent    : %.3f \n', obj.CbBudgetPercent);
           fprintf(fileID, 'Cr byte budget percent    : %.3f \n', 1 - obj.lumaBudgetPercent - obj.CbBudgetPercent);
           fprintf(fileID, 'LP gop byte budget percent: %.3f \n', obj.lpPercent);
           
           writtenBytes = 0;
           files = dir(obj.outputFolder);
           for i = 1:length(files)
               writtenBytes = writtenBytes + files(i).bytes;
           end
           bitRate = (writtenBytes / 1000 / obj.numOfFrames) * 8 * obj.frameRate;
           fprintf(fileID, 'Bytes written to file: %d (%.3f kbps) \n', writtenBytes, bitRate);
           
           fprintf(fileID, '\nTotal encoding time: %.3f sec. \n\n', toc(startTime));
           
           fclose(fileID);
       end
       
   end
   
end