classdef YUVReader < handle
    % Class for reading 8 bit YUV420 files.
    % Opens the specified file. File remains open until close method is
    % called.
    %
    % Author: F. T. Erlid
    %
    properties (SetAccess = private)
        fileName            % File name of .yuv file to read
        fileID              % fileID as returned from fopen
        frameWidth          % Width of frames
        frameHeight         % Height of frames
    end %properties
    
    methods
        
        %Constructor
        function obj = YUVReader(fileName, frameWidth, frameHeight, firstFrame)
            l = length(fileName);
            if (l < 4 || ~strcmp(fileName(l-3:l), '.yuv'))
                error('Invalid file name');
            end
            if ( frameWidth <= 0 || rem(frameWidth, 1) ~= 0 || frameHeight <= 0 || rem(frameHeight, 1) ~= 0 )
                error('Frame dimensions should be positive integers');
            end
            obj.fileName = fileName;
            obj.frameWidth = frameWidth;
            obj.frameHeight = frameHeight;
            obj.fileID = fopen(obj.fileName);
            if (fseek(obj.fileID, firstFrame * obj.frameHeight * obj.frameWidth * (3/2), 0) == -1)
                error('Failed to go to first frame');
            end
        end
        
        % Reads next frame. Returns uint8 values.
        function [luma, Cb, Cr] = readNextFrame(obj)
            if ( ~hasNextFrame(obj) )
                error('End of file is reached');
            end
            luma = fread(obj.fileID, [obj.frameWidth, obj.frameHeight], '*uint8');
            luma = luma';
            
            Cb = fread(obj.fileID, [obj.frameWidth/2, obj.frameHeight/2], '*uint8');
            Cb = Cb';
            
            Cr = fread(obj.fileID, [obj.frameWidth/2, obj.frameHeight/2], '*uint8');
            Cr = Cr';
            
        end
        
        % Reads the specified number of frames. Returns uint8 values.
        function [luma, Cb, Cr] = readNextFrames(obj, numOfFramesToRead)
           luma = uint8( zeros(obj.frameHeight, obj.frameWidth, numOfFramesToRead) );
           Cb = uint8( zeros(obj.frameHeight/2, obj.frameWidth/2, numOfFramesToRead) );
           Cr = uint8( zeros(obj.frameHeight/2, obj.frameWidth/2, numOfFramesToRead) );
           for i = 1:numOfFramesToRead
               [luma_temp, Cb_temp, Cr_temp] = readNextFrame(obj);
               luma(:, :, i) = luma_temp;
               Cb(:, :, i) = Cb_temp;
               Cr(:, :, i) = Cr_temp;
           end
        end
        
        function a = hasNextFrame(obj)
            a = abs( feof(obj.fileID) - 1 );
        end
        
        % nextFrameIndex
        %   Index of next frame (0-indexed)
        function i = nextFrameIndex(obj)
            if ( ~hasNextFrame(obj) )
                error('End of file is reached');
            end            
            i = ftell(obj.fileID) / ( obj.frameHeight * obj.frameWidth * (3/2) ); %nextFrameIndex variabel istedet? inkrementeres i readNextFrame()
        end
        
        % close
        %   Returns 0 if success, -1 otherwise
        function status = close(obj)
            status = fclose(obj.fileID);
        end
        
        % delete
        %   Makes sure file is closed on object destruction
        function delete(obj)
            fclose(obj.fileID);
        end
        
    end %methods
end