classdef YUVWriter < handle
    % Class for writing 8 bit YUV420 files.
    % Opens or creates the specified file. File remains open until close
    % method is called.
    %
    % Author: F. T. Erlid
    %
    properties (SetAccess = private)
        fileName            % File name of .yuv file to read
        fileID              % fileID as returned from fopen
        frameWidth          % Width of frames
        frameHeight         % Height of frames
    end %properties
    
    methods
        
        %Constructor
        function obj = YUVWriter(fileName, frameWidth, frameHeight)
            l = length(fileName);
            if (l < 4 || ~strcmp(fileName(l-3:l), '.yuv'))
                error('Invalid file name');
            end
            if ( frameWidth <= 0 || rem(frameWidth, 1) ~= 0 || frameHeight <= 0 || rem(frameHeight, 1) ~= 0 )
                error('Frame dimensions should be positive integers');
            end
            obj.fileName = fileName;
            obj.frameWidth = frameWidth;
            obj.frameHeight = frameHeight;
            obj.fileID = fopen(obj.fileName, 'w');
        end
        
        % Writes next frame. Returns number of bytes written.
        function writtenBytes = writeNextFrame(obj, luma, Cb, Cr)
            if ( ~strcmp(class(luma), 'uint8') || ~strcmp(class(Cb), 'uint8') || ~strcmp(class(Cr), 'uint8') )
                error('Luma, Cb and Cr values must be of class uint8');
            end
            if ( length(size(luma)) ~= 2 || length(size(Cb)) ~= 2 || length(size(Cr)) ~= 2)
                error('Only single frames is accepted as input. Use writeNextFrames() instead');
            end
            dim_luma = size(luma);
            dim_Cb = size(Cb);
            dim_Cr = size(Cr);
            if ( ~(dim_luma(1) == obj.frameHeight && dim_luma(2) == obj.frameWidth) || ~(dim_Cb(1) == obj.frameHeight/2 && dim_Cb(2) == obj.frameWidth/2) )
                error('Input dimensions are incorrect');
            end
            if ( ~(dim_Cb(1) == dim_Cr(1)) || ~(dim_Cb(2) == dim_Cr(2)) )
                error('Chroma components are not of equal size');
            end
            writtenBytes = fwrite(obj.fileID, luma', 'uint8');
            writtenBytes = writtenBytes + fwrite(obj.fileID, Cb', 'uint8');
            writtenBytes = writtenBytes + fwrite(obj.fileID, Cr', 'uint8');
        end
        
        % Writes the specified number of frames. Returns number of bytes written.
        function writtenBytes = writeNextFrames(obj, luma, Cb, Cr)
            numOfFramesToWrite = size(luma);
            numOfFramesToWrite = numOfFramesToWrite(3);
            
            writtenBytes = 0;
            for i = 1:numOfFramesToWrite
                writtenBytes = writtenBytes + writeNextFrame( obj, luma(:,:,i), Cb(:,:,i), Cr(:,:,i) );
            end
        end
        
        % close
        %   Returns 0 if success, -1 otherwise
        function status = close(obj)
            status = fclose(obj.fileID);
        end
        
        % delete
        %   Makes sure file is closed on object destruction
        function delete(obj)
            fclose(obj.fileID);
        end
        
    end %methods
end