function writtenBytes = jp2k_gop_encoder(luma_subbands, Cb_subbands, Cr_subbands, GOPnumber, byteBudget, outputFolder, varargin)
% Subband are written with JasPer 1.900.1, and stored as 
% jpc/j2c (Raw JPEG 2000 code stream).

if (ndims(luma_subbands) == 3)
    GOP_size = size(luma_subbands);
    GOP_size = GOP_size(3);
else
    GOP_size = 1;
end

if(byteBudget > 0)
    if (nargin == 6)
        lumaGopByteBudget = byteBudget * 0.85; % Assign 85% of budget to luma
        CbGopByteBudget = byteBudget * 0.1; % Assign 10% of budget to Cb
        CrGopByteBudget = byteBudget - lumaGopByteBudget - CbGopByteBudget; % Assign rest of budget to Cr
        lpPercent = 0.25;
    elseif (varargin{1} > 0 && varargin{2} > 0 && varargin{3} > 0)
        lumaGopByteBudget = byteBudget * varargin{1}; % Assign varargin{1}% of budget to luma
        CbGopByteBudget = byteBudget * varargin{2}; % Assign varargin{2}% of budget to Cb
        CrGopByteBudget = byteBudget - lumaGopByteBudget - CbGopByteBudget; % Assign rest of budget to Cr
        lpPercent = varargin{3};
    end
    
    if (GOP_size > 1)
        lumaLPSubbandByteBudget = lumaGopByteBudget * lpPercent;
        lumaHPSubbandByteBudget = (lumaGopByteBudget - lumaLPSubbandByteBudget) / (GOP_size - 1);        
        
        CbLPSubbandByteBudget = CbGopByteBudget * lpPercent;
        CbHPSubbandByteBudget = (CbGopByteBudget - CbLPSubbandByteBudget) / (GOP_size - 1);
        
        CrLPSubbandByteBudget = CrGopByteBudget * lpPercent;
        CrHPSubbandByteBudget = (CrGopByteBudget - CrLPSubbandByteBudget) / (GOP_size - 1);
    else
        lumaLPSubbandByteBudget = lumaGopByteBudget;        
        CbLPSubbandByteBudget = CbGopByteBudget;
        CrLPSubbandByteBudget = CrGopByteBudget;
    end
end

filePrefix = [outputFolder '/gop_' int2str(GOPnumber) '_sub_'];
filePostfix = 'j2c';

% Write lowpass frames in uint8 format
writeJp2k(luma_subbands(:,:,1), outputFolder, lumaLPSubbandByteBudget, [filePrefix '1_luma.' filePostfix], ConstantProperties.NO_JP2K_SPATIAL_LEVELS_LP_LUMA);
writeJp2k(Cb_subbands(:,:,1), outputFolder, CbLPSubbandByteBudget, [filePrefix '1_cb.' filePostfix], ConstantProperties.NO_JP2K_SPATIAL_LEVELS_LP_CHROMA);
writeJp2k(Cr_subbands(:,:,1), outputFolder, CrLPSubbandByteBudget, [filePrefix '1_cr.' filePostfix], ConstantProperties.NO_JP2K_SPATIAL_LEVELS_LP_CHROMA);

% Write highpass frames.
for i = 2:GOP_size
    %temp_luma = subbandCutoff( luma_subbands(:,:,i) );
    [temp_luma, sf] = subbandScaling( luma_subbands(:,:,i) );
    writeJp2k(temp_luma, outputFolder, lumaHPSubbandByteBudget, [filePrefix int2str(i) '_luma.' filePostfix], ConstantProperties.NO_JP2K_SPATIAL_LEVELS_HP_LUMA, sf);
    
    %temp_Cb = subbandCutoff( Cb_subbands(:,:,i) );
    [temp_Cb, sf2] = subbandScaling( Cb_subbands(:,:,i) );
    writeJp2k(temp_Cb, outputFolder, CbHPSubbandByteBudget, [filePrefix int2str(i) '_cb.' filePostfix], ConstantProperties.NO_JP2K_SPATIAL_LEVELS_HP_CHROMA, sf2);
    
    %temp_Cr = subbandCutoff( Cr_subbands(:,:,i) );
    [temp_Cr, sf3] = subbandScaling( Cr_subbands(:,:,i) );
    writeJp2k(temp_Cr, outputFolder, CrHPSubbandByteBudget, [filePrefix int2str(i) '_cr.' filePostfix], ConstantProperties.NO_JP2K_SPATIAL_LEVELS_HP_CHROMA, sf3);
end

% Determine how many bytes was written
writtenBytes = 0;
files = dir( [filePrefix '*'] );
for i = 1:length(files)
    writtenBytes = writtenBytes + files(i).bytes;
end

end


function y = subbandCutoff(subband)
% HP frames have values in range [-255, 255]. First, add 127 to all values
% to make range [-128, 382], then set values < 0 to zero and values above
% 255 to 255. This avoids having to store hp subbands as 16bit jp2k.

y = subband + 127;
y(y < 0) = 0;
y(y > 255) = 255;
end

function [y, scalingFactor] = subbandScaling(subband)
% HP frames have values in range [-255, 255]. First, find scaling factor in
% range [0.5, 1] with 0.1 step size. Scale values, add 127 and round to
% nearest integer in range [0, 255].

scalingFactor = ceil(( 127 / max(max(abs(subband))) ) * 10) / 10;
if (scalingFactor > 1)
    scalingFactor = 1;
end
y = subband * scalingFactor;
y = y + 127;
y = round(y);
y(y < 0) = 0;
y(y > 255) = 255;
end


function writeJp2k(data, outputFolder, subbandByteBudget, fileName, noSpatialLevels, varargin)
sfBytes = 0;
if (nargin == 6)
    scalingFactor_temp = varargin{1};
    sf = 0;
    switch scalingFactor_temp
        case 0.5
            sf = 0;
        case 0.6
            sf = 1;
        case 0.7
            sf = 2;
        case 0.8
            sf = 3;
        case 0.9
            sf = 4;
        case 1
            sf = 5;
        otherwise
            error('Bad scaling factor');
    end            
    sfBytes = 0.375; % 3 bits needed to write scaling factor
end
redundantMainHeaderBytes = 98;
redundantTileHeaderBytes = 14;
dim = size(data);
compressionRate = (subbandByteBudget + redundantMainHeaderBytes + redundantTileHeaderBytes - sfBytes) / ( dim(1) * dim(2) );

tempFile = [outputFolder '/temp.pgm'];
imwrite( uint8(data), tempFile, 'pgm');

runStringPrefix = ['jasper -f ' tempFile ' -t pnm -T jpc -O mode=real '];
rateString = '';

if (compressionRate > 0)
    rateString = [' -O rate=' num2str(compressionRate) ' '];
end

system( [runStringPrefix '-F ' fileName rateString ' -O numrlvls=' num2str(noSpatialLevels)] );

% Remove redundant information from Main header
fileIDin = fopen(fileName);
fileIDout = fopen([fileName '.bin'], 'w');
fseek(fileIDin, redundantMainHeaderBytes, 'bof'); % Seek past SOC, SIZ, COM (comment generated by JasPer) and COD segments, and QCD marker

%Write scaling factor if needed
if (nargin == 6)
    fwrite(fileIDout, sf, 'ubit3');
end

%Write QCD segment (without QCD marker)
qcdSize = fread(fileIDin, 1, 'uint16', 0, 'b');
fwrite(fileIDout, qcdSize, 'uint16', 0, 'b');
fwrite(fileIDout, fread(fileIDin, qcdSize-2, 'uint8', 0, 'b'), 'uint8', 0, 'b');

%Seek past tile header (SOT, Lsot, i_T, L_TP, i_TP, N_TP,t and SOD)
fseek(fileIDin, redundantTileHeaderBytes, 'cof');

fwrite(fileIDout, fread(fileIDin));

fclose(fileIDin);
fclose(fileIDout);
delete(fileName);

delete(tempFile);
    
end