function [subband_frames, mvx, mvy] = mctf_gop_encoder(frames, numOfTemporalLevels, filterType, blockSize, searchRange, searchType, varargin)
%GOP encoder
%varargin{1} = idr
%varargin{2} = OBMC (boolean to decide whether to use OBMC or not)

OBMC = 1;
if (nargin == 8) %husk: >= 8 hvis flere
    OBMC = varargin{2};
end

% Check parameters
if( ndims(frames) ~= 3 )
    error('Frames parameter must contain 2 or more frames');
end

[frame_height, frame_width, GOP_size] = size(frames);

if(GOP_size ~= 2^numOfTemporalLevels)
    error('The number of frames to encode must equal 2^numOfTemporalLevels');
end
if ( rem(blockSize, 1) ~= 0 )
    error('blockSize parameter must be integer valued');
end
if ( mod(frame_height, blockSize) ~= 0 || mod(frame_width, blockSize) ~= 0)
    error('Frame dimensions not divisible by Block size');
end

if( strcmpi(filterType, 'haar') ) %NB: out of date after OBMC inclusion
        
    mvx = zeros(frame_height / blockSize, frame_width / blockSize, GOP_size - 1);
    mvy = zeros(frame_height / blockSize, frame_width / blockSize, GOP_size - 1);
    subband_frames = zeros(frame_height, frame_width, GOP_size);
    
    lowpass_frames = frames;
    level_index = numOfTemporalLevels;
    searchRange = searchRange / 2^(numOfTemporalLevels-1); %Search range for first temporal level
    while( level_index > 0)
        lowpass_frames_temp = zeros(frame_height, frame_width, 2^(level_index-1));
        for i = 1:2:2^level_index
            frame_0 = lowpass_frames(:, :, i);
            frame_1 = lowpass_frames(:, :, i+1);
            
            [mvx_temp, mvy_temp] = block_matching_motion_estimation(frame_1, frame_0, blockSize, searchRange, searchType);
            [l_temp, h_temp] = HaarMCTF.analysis(frame_0, frame_1, mvx_temp, mvy_temp, blockSize);
            
            mvx(:, :, 2^(level_index-1)+ceil(i/2)-1) = mvx_temp;
            mvy(:, :, 2^(level_index-1)+ceil(i/2)-1) = mvy_temp;
            
            lowpass_frames_temp(:, :, ceil(i/2)) = l_temp;
            subband_frames(:, :, 2^(level_index-1)+ceil(i/2)) = h_temp;
        end
        level_index = level_index - 1;
        searchRange = searchRange * 2; %Double search range for next level
        lowpass_frames = lowpass_frames_temp;
    end
    subband_frames(:, :, 1) = lowpass_frames(:, :, 1);
    
    
elseif( strcmpi(filterType, '1/3') )
    
    if (~OBMC)
        mvx = zeros(frame_height / blockSize, frame_width / blockSize, 2^(numOfTemporalLevels+1) - 2);
        mvy = zeros(frame_height / blockSize, frame_width / blockSize, 2^(numOfTemporalLevels+1) - 2);
    else %OBMC
        mvx = zeros((frame_height / (blockSize/2)-1), (frame_width / (blockSize/2)-1), 2^(numOfTemporalLevels+1) - 2);
        mvy = zeros((frame_height / (blockSize/2)-1), (frame_width / (blockSize/2)-1), 2^(numOfTemporalLevels+1) - 2);
    end

    subband_frames = zeros(frame_height, frame_width, GOP_size);
    
    lowpass_frames = zeros(frame_height, frame_width, GOP_size + 1);
    lowpass_frames(:,:,1) = varargin{1};
    for i = 1:GOP_size
        lowpass_frames(:,:,i+1) = frames(:,:,i);
    end
    level_index = numOfTemporalLevels;
    searchRange = searchRange / 2^(numOfTemporalLevels-1); %Search range for first temporal level
    while( level_index > 0)        
        for i = 1:2:2^level_index
            frame_0 = lowpass_frames(:, :, i);
            frame_1 = lowpass_frames(:, :, i+1);
            frame_2 = lowpass_frames(:, :, i+2);
            
            if (~OBMC)
                [mvx_f_temp, mvy_f_temp] = block_matching_motion_estimation(frame_1, frame_0, blockSize, searchRange, searchType);
                [mvx_b_temp, mvy_b_temp] = block_matching_motion_estimation(frame_1, frame_2, blockSize, searchRange, searchType);
            else %OBMC
                [mvx_f_temp, mvy_f_temp] = obmc(frame_1, frame_0, blockSize, searchRange);
                [mvx_b_temp, mvy_b_temp] = obmc(frame_1, frame_2, blockSize, searchRange);
            end

            h_temp = OneThreeMCTF.analysis(frame_0, frame_1, frame_2, mvx_f_temp, mvy_f_temp, mvx_b_temp, mvy_b_temp, blockSize, OBMC);
            
            mvx(:, :, 2^(level_index)-1+(i-1)) = mvx_f_temp;
            mvx(:, :, 2^(level_index)+(i-1)) = mvx_b_temp;
            
            mvy(:, :, 2^(level_index)-1+(i-1)) = mvy_f_temp;
            mvy(:, :, 2^(level_index)+(i-1)) = mvy_b_temp;
            
            subband_frames(:, :, 2^(level_index-1)+ceil(i/2)) = h_temp;
        end      
        lowpass_frames = lowpass_frames(:,:,1:2:2^level_index+1);
        level_index = level_index - 1;  
        searchRange = searchRange * 2; %Double search range for next level
    end
    subband_frames(:, :, 1) = frames(:, :, GOP_size);
end

end